# -*- coding: utf-8 -*-
#
# Copyright (c) 2011 Daniel Gerber.
#
#This program is free software: you can redistribute it and/or modify
#it under the terms of the GNU General Public License as published by
#the Free Software Foundation, either version 3 of the License, or
#(at your option) any later version.
#
#This program is distributed in the hope that it will be useful,
#but WITHOUT ANY WARRANTY; without even the implied warranty of
#MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#GNU General Public License for more details.
#
#You should have received a copy of the GNU General Public License
#along with this program. If not, see <http://www.gnu.org/licenses/>.

# Modified by Ivan Herman, 2012, to produce a JSON-LD serialization: http://json-ld.org/spec/latest/json-ld-syntax
"""
 $Id: jsonserializer.py,v 1.4 2012-12-05 19:18:17 ivan Exp $
"""
import sys

if sys.version_info[1] >= 7 :
	from collections import OrderedDict
else :
	from pyRdfaExtras.extras.odict import odict as OrderedDict

from rdflib.serializer import Serializer
from rdflib.term import URIRef, BNode, Literal
from rdflib	import RDF  as ns_rdf
from rdflib	import RDFS as ns_rdfs

RDFA_VOCAB = URIRef("http://www.w3.org/ns/rdfa#usesVocabulary")

class JsonSerializer(Serializer):
	__doc__ = __doc__
	# List of predicates that have a special usage and should not appear as part of the coerced list in the context
	non_coerced_predicates = [ ns_rdf["type"], ns_rdf["first"], ns_rdf["rest"] ]
	automatic_datatypes    = [ URIRef("http://www.w3.org/2001/XMLSchema#integer"),
							   URIRef("http://www.w3.org/2001/XMLSchema#double"),
							   URIRef("http://www.w3.org/2001/XMLSchema#boolean")
							 ]

	def __init__(self, graph):
		self.graph          = graph
		# Coerce structure. Is filled through _initialize_predicates()
		self.coerce         = OrderedDict()
		self.prefix_map     = PrefixMap(self.graph.namespaces())
		# List of subjects that should appear on top. Initialized to all subjects,
		# trimmed in _initialize_subjects()
		self.top_subjects   = set([ s for s in self.graph.subjects() ])
		# dictionary mapping subjects to JSON structures
		self.all_subjects   = OrderedDict()
		# Set of subjects that may appear as intermediate objects in a chain
		self.chain_links    = set()
		# Set of predicates whose range in this graph are URI References or Blank Nodes
		self.uri_predicates = set()
		# Single vocabulary, as generated by an RDFa processor
		self.vocab          = None
		# The 'owner' or the vocabulary, ie, the subject of the vocab setting triple (if there is only one)
		self.vocab_owner    = None
		# Vocabulary terms, ie, those terms that can be represented as being part of the vocab
		self.vocabulary_terms = {}
		# Lists
		self.lists          = {}
		# This is to keep the ancestors quiet
		self.base = None

	def serialize(self, stream, base=None, encoding='utf-8', **kwds):
		""" Generic entry point for the serialization, as used by RDFLib.
		    The serializer uses order preserving dictionaries to keep the right/expected order within the JSON-LD output, too. 
		"""
		if encoding == None : encoding = 'utf-8'
		# Create the dictionary to be serialized
		d = self._build(base=base, **kwds)
		
		if sys.version_info[1] >= 6 :
			import json
			s = json.dumps(d, ensure_ascii = False, indent=4)
		else :
			import simplejson
			s = simplejson.dumps(d, ensure_ascii = False, indent=4)

		try :
			stream.write(s)
		except UnicodeEncodeError :
			# This stuff occurs sometimes, and I am not sure why:-()
			stream.write(s.encode(encoding))

	def _build(self, base=None, prefix_map=None, encode_literal=None, **kwds):
		"""Returns an ordered dict to serialize. The base is, in fact, unused, it is only here
		because it appears in the rest of the RDFLib call structure..."""
		if encode_literal:
			assert callable(encode_literal)
			self._encode_literal = encode_literal
			
		if prefix_map:
			self.prefix_map.update(prefix_map)

		# Subjects: find the possible 'top level' subjects, possible links; result is a dictionary referring to the
		# json objects to produce
		self._initialize_subjects()
		# Predicates: find the predicates that can safely be coerced as producing URI-s
		self._initialize_predicates()
		# Name tells it all: find possible list structures in the graph and treat them separately
		self._initialize_lists()
		# Find out if the @vocab has been used in the graph; if so, and there is only one, the term usage is
		# reproduced in the JSON LD output
		self._rdfa_vocabulary_usage()

		# Fill in the content of the json objects 
		for s in self.all_subjects.keys() :
			# List headers are treated specially
			if s in self.lists : continue

			# Get the subject structure
			subj = self.all_subjects[s]
			# Get the possible types, and encode them through the special json-ld syntax
			types = [ t for t in self.graph.objects(s,ns_rdf["type"]) ]
			if len(types) == 1 :
				subj["@type"] = self._get_node_ref(t)
			elif len(types) > 1 :
				subj["@type"] = [ self._get_node_ref(t) for t in types ]

			# Get the other properties and their objects
			for p in self.graph.predicates(s) :
				# Types have already been taken care of
				if p == ns_rdf["type"] : continue
	
				pobj = self._predicate(p)
				objs = [ o for o in self.graph.objects(s,p) ]
				
				# The cardinality of objs makes a difference in the output format...
				if len(objs) == 0 :
					# Should not happen, though
					continue
				elif len(objs) == 1 :
					subj[pobj] = self._object(objs[0], s, p)
				else :
					subj[pobj] = [ self._object(ob, s, p) for ob in objs ]
					
		#######################################################################
		# Yet another beautification: if a top level object has no parent and is a blank node,
		# the @id is unnecessary.
		for s in self.top_subjects :
			if len([ p for (p,x,y) in self.graph.triples((None, None, s)) ]) == 0 :
				if isinstance(s, BNode) :
					self.all_subjects[s].pop("@id",None)
					
		#######################################################################
		# Put all together now in the top level object, ie, the one that produces the output
		_json_obj = OrderedDict()

		# Add the @context part, if needed
		if len(self.prefix_map.used_keys) > 0 or len(self.vocabulary_terms) > 0 or len(self.uri_predicates) > 0 :
			context = OrderedDict()
			predicate_handled = set()
			
			# Add the context for CURIE-s
			# If the predicate has been identified as producing URI references only, that is taken care of here
			for k,v in self.graph.namespaces() :
				if k in self.prefix_map.used_keys :
					predicate_handled.add(v)
					if v in self.uri_predicates :
						typ = OrderedDict()
						typ['@id']   = "%s" % v
						typ['@type'] = '@id'
						context[k] = typ
					else :
						context[k] = "%s" % v
				
			# Add the context for predicates that originate from an RDFa @vocab construct			
			# If the predicate has been identified as producing URI references only, that is taken care of here
			for k,v in self.vocabulary_terms.items() :
				predicate_handled.add(v)
				if v in self.uri_predicates :
					typ = OrderedDict()
					typ['@id']   = "%s" % v
					typ['@type'] = '@id'
					context[k] = typ
				else :
					context[k] = "%s" % v

			# Some predicates might have been used as full URI-s, but should still be added
			# for type coercion if they produce URI references only
			for p in self.uri_predicates :
				if p not in predicate_handled :
					typ = OrderedDict()
					typ['@type'] = '@id'
					cp = self.prefix_map.shrink(p)
					context[cp if cp != None else p] = typ
					
			# Context is done
			_json_obj["@context"] = context

		# Add the top level objects; the content of these have been filled by the previous steps
		# There is a big difference on whether there are several top level object or not; in the former
		# case the special JSON-LD "@graph" keyword has to be used as a key
		if len(self.top_subjects) == 1 :
			subj = self.all_subjects[self.top_subjects.pop()]
			for k in subj.keys() :
				_json_obj[k] = subj[k]
		elif len(self.top_subjects) > 1 :
			_json_obj["@graph"] = [ self.all_subjects[s] for s in self.top_subjects ]

		return _json_obj
	
	def _initialize_subjects(self) :
		"""Collect the various subject categories: initial top level list,
		chain links; plus a dictionary mapping subjects to their json structure.
		"""
		# First get all the subjects in one place
		# all subjects are represented by a json object, store those, too (though intially empty, but the
		# structure is then in place)
		# The initial list of subjects is also stored in the top_subject array,
		# ie, the list of those subjects that will appear on top of the serialized json output
		for s in self.top_subjects :
			js_struct = OrderedDict()
			# This may have to be refined to remove blank nodes in a chain link...
			js_struct["@id"] = self._get_node_ref(s)
			self.all_subjects[s] = js_struct
			
		# Now the chains have to be found
		for s in self.top_subjects :
			# see the number of parents, ie, the number of other subjects for which this subject appear as an object.
			# if the number is exactly one, this means this subject may appear in a chaining structure
			if len([ p for (p,x,y) in self.graph.triples((None, None, s)) ]) == 1 :
				self.chain_links.add(s)
				
	def _initialize_predicates(self) :
		"""Collect all those predicates whose objects are URIRefs only. This collection can be used
		to simplify the output by using JSON-LD's coercion facility.
		"""
		for p in set([ p for p in self.graph.predicates() ]) :
			if p not in self.non_coerced_predicates and True not in [isinstance(o, Literal) for s,x,o in self.graph.triples((None,p,None))] :
				self.uri_predicates.add(p)
			
	def _initialize_lists(self) :
		# Find and create the list structures
		def get_heads(l) :
			retval = []
			nl = l
			while True :				
				h = self.graph.value(nl,ns_rdf["rest"])
				if h != ns_rdf["nil"] :
					retval.append(h)
					nl = h
				else :
					break
			return retval
		
		for s in self.all_subjects.keys() :
			# See if this is a possible list in the first place
			if self.graph.value(s, ns_rdf["first"]) != None and self.graph.value(s, ns_rdf["rest"]) != None :
				if len([p for p,x,y in self.graph.triples((None,ns_rdf["rest"],s))]) == 0 :
					# Yep, this is the head of a list!
					# Let us collect the list. First the array of list content:
					content = [ c for c in self.graph.items(s) ]
					heads   = get_heads(s)
					# not a clean list...
					if False in [ isinstance(h, BNode) and h in self.chain_links for h in heads ] : break
					self.lists[s] = (content, heads)
		# Now we will have to massage the lists and the subject informations...
		for s in self.lists :
			content, heads = self.lists[s]
			
			# First the corresponding structure for the list head should change, turning it into the JSON-LD
			# abbreviation for lists
			lst = OrderedDict()
			lst["@list"] = [ self._object(c, None, None) for c in content ]
			self.all_subjects[s] = lst
			
			# The heads, ie, the list building blocks, should be removed from further processing
			for h in heads :
				self.all_subjects.pop(h, None)
				self.top_subjects.discard(h)

	def _get_node_ref(self, ident):
		"Returns the property name / local identifier for this node."
		if isinstance(ident, URIRef):
			if self.vocab and ident.startswith(self.vocab) :
				term = ident.replace(self.vocab,'',1)
				# Let us not create an empty term if the vocabulary URI is referred to anywhere...
				if term == "" : return ident
				self.vocabulary_terms[term] = ident
				return term
			else :
				return self.prefix_map.shrink(ident) or self.relativize(ident)
		elif isinstance(ident, BNode):
			return "_:%s" % ident
		raise TypeError('Expected URIRef or BNode instance, got %r' % ident)

	_predicate = _get_node_ref
	
	def _object(self, ident, parent, predicate):
		"Returns a json-serializable."
		if isinstance(ident, Literal):
			return self._encode_literal(ident)
		else :
			# This is a bnode or a URI ref. We have to see if it is a possible chain link
			if ident in self.chain_links :
				# Yep!
				# This object should be removed from the list of top level subjects
				self.top_subjects.discard(ident)
				# The parent must me removed from the chain_links, if there...
				if parent : self.chain_links.discard(parent)
				# The json structure of this object should be linked from the parent; the
				# content will be filled later
				jsubj = self.all_subjects[ident]
				# If this happens to be a BNode, this is the case when the BNode id is unnecessary
				if isinstance(ident, BNode) :
					jsubj.pop("@id",None)			
				return jsubj
			else :
				# no, this is either a leaf or a subject with multiple parents
				if predicate and predicate in self.uri_predicates :
					# The coercion rules will be added to the context for this predicate
					if isinstance(ident, BNode) :
						return "_:%s" % ident
					else :
						return self._get_node_ref(ident)
				else :
					retval = OrderedDict()
					if isinstance(ident, BNode) :
						retval["@id"] = "_:%s" % ident
					else :
						retval["@id"] = self._get_node_ref(ident)
					return retval
	
	def _encode_literal(self, literal):
		"""Produce either a plain string as a literal, or a literal with datatype or language"""
		if literal.datatype != None :
			retval = OrderedDict()
			if literal.datatype in 	JsonSerializer.automatic_datatypes :
				try :
					return literal.toPython()
				except :
					retval["@value"] = literal
					retval["@type"] = self._get_node_ref(literal.datatype)
			else :
				retval["@value"] = literal
				retval["@type"] = self._get_node_ref(literal.datatype)
			return retval
		elif literal.language != None and literal.language != "" :
			retval = OrderedDict()
			retval["@value"] = literal
			retval["@language"] = literal.language
			return retval
		else :
			return literal

	# This part is very specific to RDFa usage!
	def _rdfa_vocabulary_usage(self) :
		# See if a vocabulary has been used at all and, if yes, whether there is only one
		vocabs = [ (s,p,o) for s,p,o in self.graph.triples((None, RDFA_VOCAB, None)) ]
		# There may be a double management of vocab and namespace (schema.org is a typical case). The vocab approach
		# should prevail, otherwise problems occur in the @context of the generated JSON-LD with the keys
		for v in vocabs :
			self.prefix_map.remove_vocab(v[2])
			
		# if there is no vocab, or if there is more than one, there is nothing we can do to beautify
		# (in the multiple vocabulary case, there may be identical terms used in both vocabularies, and that can lead
		# to a mess)
		if len(vocabs) == 1 :
			self.vocab       = vocabs[0][2]
			self.vocab_owner = vocabs[0][0]
			
			# I had this code, part of a beautifying step that removed the RDFa triple on vocabularies if there was only
			# one. I think that is not kosher, so I decided to comment it out.
			#parents  = [ p for (p,x,y) in self.graph.triples((None, None, self.vocab_owner)) ]
			#children = [ p for (x,y,p) in self.graph.triples((self.vocab_owner, None, None)) ]
			## See if this is the only apperance of the vocab owner; if so, it should be removed from further processing
			#if len(parents) == 0 and len(children) == 1:
			#	# the subject should be removed
			#	self.top_subjects.discard(self.vocab_owner)

class PrefixMap(dict):
	"""A mapping of prefixes to URIs."""

	def __init__(self, parent=None, *args, **kwds):
		dict.__init__(self, *args, **kwds)
		self.parent     = parent and dict(parent) or None
		self.used_keys  = set()
		self.vocab_uris = set()
		
	def remove_vocab(self, uriref) :
		self.vocab_uris.add(uriref)
		
	def shrink(self, uriref):
		"Returns a CURIE or None."
		for pfx, ns in self.iteritems():
			if ns not in self.vocab_uris and uriref.startswith(ns):
				self.used_keys.add(pfx)
				return '%s:%s' % (pfx, uriref.replace(ns,'',1))
		if self.parent:
			for pfx, ns in self.parent.iteritems():
				if ns not in self.vocab_uris and uriref.startswith(ns):
					self[pfx] = ns
					self.used_keys.add(pfx)
					return '%s:%s' % (pfx, uriref.replace(ns,'',1))

	def resolve(self, curie):
		"Returns an URIRef or None."
		prefix, s, relative_ref = curie.partition(':')
		if prefix in self:
			return URIRef(self[prefix] + relative_ref)
		elif self.parent and prefix in self.parent:
			return URIRef(self.parent[prefix] + relative_ref)



"""

XML Writer to be used by the L{PrettyXMLSerializer} module. This is a slightly modified version of RDFLib's 
original XML Writer module, the changes are essentially aesthetic in nature (removing the unnecessary CDATA sections, 
using double quotes everywhere for attributes (both for namespaces and attributes in general).

"""

import codecs
from xml.sax.saxutils import quoteattr, escape

class XMLWriter(object):
	def __init__(self, stream, namespace_manager, encoding=None, decl=1):
		encoding = encoding or 'utf-8'
		encoder, decoder, stream_reader, stream_writer = codecs.lookup(encoding)
		self.stream = stream = stream_writer(stream)
		if decl:
			stream.write('<?xml version="1.0" encoding="%s"?>' % encoding)
		self.element_stack = []
		self.nm = namespace_manager
		self.closed = True

	def __get_indent(self):
		return "  " * len(self.element_stack)
	indent = property(__get_indent)

	def __close_start_tag(self):
		if not self.closed: # TODO:
			self.closed = True
			self.stream.write(">")

	def push(self, uri):
		nm = self.nm
		self.__close_start_tag()
		write = self.stream.write
		write("\n")
		write(self.indent)
		write("<%s" % nm.qname(uri))
		self.element_stack.append(uri)
		self.closed = False
		self.parent = False

	def pop(self, uri=None):
		top = self.element_stack.pop()
		if uri:
			assert uri==top
		write = self.stream.write
		if not self.closed:
			self.closed = True
			write("/>")
		else:
			if self.parent:
				write("\n")
				write(self.indent)
			write("</%s>" % self.nm.qname(uri))
		self.parent = True

	def namespaces(self, namespaces):
		write = self.stream.write
		write("\n")
		for prefix, namespace in namespaces:
			if prefix:
				write('  xmlns:%s="%s"\n' % (prefix, namespace))
			else:
				write('  xmlns="%s"\n' % namespace)

	def attribute(self, uri, value):
		write = self.stream.write
		write(" %s=%s" % (self.nm.qname(uri), quoteattr(value)))

	def text(self, text, to_escape):
		self.__close_start_tag()
		if to_escape :
			self.stream.write(escape(text))
		else :
			self.stream.write(text)
#		if "<" in text and ">" in text and not "]]>" in text:
#			#self.stream.write("<![CDATA[")
#			self.stream.write(text)
#			#self.stream.write("]]>")
#		else:
#			self.stream.write(escape(text))



#!/usr/local/bin/python3
# -*- coding: utf-8 -*-
# Maintainer: Ivan Herman <ivan@w3.org>

"""
Possible CGI entry point for the RDFa 1.1 package.

This version is adapted to the particualarities of the W3C setup as well as my own machine for Python paths


@author: U{Ivan Herman<a href="http://www.w3.org/People/Ivan/">}
@license: This software is available for use under the
U{W3C® SOFTWARE NOTICE AND LICENSE<href="http://www.w3.org/Consortium/Legal/2002/copyright-software-20021231">}
@contact: Ivan Herman, ivan@w3.org
"""

"""
$Id: RDFa.py,v 1.27 2018/05/23 08:57:19 carcone Exp $
"""

__version__ = "4.0.0"
import cgi
import cgitb
import sys, os
#import StringIO
#cgi.print_environ()

if sys.platform == "darwin" :
	# this is my local machine
	sys.path.insert(0,'/Users/ivan/Library/Python')
	sys.path.insert(0,'/Users/ivan/Library/Python/RDFa')
	os.environ['PyRdfaCacheDir'] = '/Users/ivan/.pyrdfa-cache'
	cgitb.enable()

else :
	# This will have to be updated for the Python3 installation!!!
	# webencodings pip3 should also be done!!!
	sys.path.insert(0,"/usr/lib/python2.7/dist-packages")
	sys.path.insert(0,'/home/ivan/lib/python')
	os.environ['PyRdfaCacheDir'] = '/usr/local/apache/cgi/cgi-bin-other/RDFa/data-local'
	cgitb.enable(display=0, logdir="/home/nobody/tracebacks/")

from pyRdfa import processURI, RDFaError

# Register the RDFa JSON-LD serializer; for some reasons installing via pip did not work
from rdflib.plugin import register, Serializer
register('json', Serializer, 'rdflib_jsonld.serializer', 'JsonLDSerializer')

def err_message(msg) :
	from cleanhtml import clean_print
	print('Content-type: text/html; charset=utf-8')
	print('Status: 400 Invalid Input')
	print()
	print("<html>")
	print("<head>")
	print("<title>Error in RDFa processing</title>")
	print("</head><body>")
	print("<h1>Error in distilling RDFa</h1>")
	print("<p>")
	clean_print("pyRdfa cannot process this URI: %s", uri)
	print("</p>")
	if len(msg) != 0 :
		print("<p>")
		clean_print(msg)
		print("</p>")
	print("</body>")
	print("</html>")
	sys.exit(1)


def brett_test(uri) :

	if not sys.platform == "darwin" :
		from checkremote import check_url_safety, UnsupportedResourceError
		from urllib2 import HTTPError, URLError
		try:
			check_url_safety(uri)
		except HTTPError as e:
			err_message('HTTP Error with the error code: %s and the error message: "%s"' (e.code, e.reason))
		except URLError as e:
			err_message('URL Error with the error message: "%s"' % e.reason)
		except UnsupportedResourceError as e:
			msg = e.args[0] + ": " + e.args[1]
			err_message('Unsupported Resource Error with the error message "%s"' % msg)
		except Exception as e:
			l = len(e.args)
			msg = "" if l == 0 else (e.args[0] if l == 1 else e.args)
			err_message('Exception raised: "%s"' % msg)

#
# to make this thing exist...
uri  = ""
form = cgi.FieldStorage()
# First see if an upload is present in the form
if "uploaded" in form and form["uploaded"].file :
	uri = "uploaded:"
elif "text" in form and form["text"].value != None and len(form["text"].value.strip()) != 0 :
	uri  = "text:"
else :
	if not "uri" in form :
		print('Content-type: text/html; charset=utf-8')
		print('Status: 400 Invalid Input')
		print()
		print("<html>")
		print("<head>")
		print("<title>Error in RDFa processing</title>")
		print("</head><body>")
		print("<h1>Error in distilling RDFa</h1>")
		print("<p>No URI has been specified</p>")
		print("</body>")
		print("</html>")
		sys.exit(1)

	try :
		uri = form.getfirst("uri")
	except :
		print('Content-type: text/html; charset=utf-8')
		print('Status: 400 Invalid Input')
		print()
		print("<html>")
		print("<head>")
		print("<title>Error in RDFa processing</title>")
		print("</head><body>")
		print("<h1>Error in distilling RDFa</h1>")
		print("<p>No URI has been specified</p>")
		print("</body>")
		print("</html>")
		sys.exit(1)

if "validate" in form  :
	from rdfavalidator import validateURI
	if not (uri == 'text:' or uri == 'uploaded:') :
		brett_test(uri)
	print('Content-Type: text/html; charset=utf-8')
	print()
	print(validateURI(uri, form))
else :
	try :
		# Thanks to Sergio and Diego for the idea and code for the referer branch
		if uri == "referer" :
			uri    = os.getenv('HTTP_REFERER')
			if uri is None:
				newuri = "http://www.w3.org/2012/pyRdfa/no_referer.html"
			else:
				brett_test(uri)
				newuri = "http://www.w3.org/2012/pyRdfa/extract?uri=" + uri
			print("Status: 307 Moved Temporarily")
			print("Location: " + newuri)
			print()
		else :
			# last point of check: use Brett's script to check the validity of the URI
			if not (uri == 'text:' or uri == 'uploaded:') :
				brett_test(uri)

			if "format" in form.keys() :
				format = form.getfirst("format")
			else :
				format = "turtle"
			retval = processURI(uri, format, form)
			print(retval)
	except Exception as e :
		l = len(e.args)
		msg = "" if l == 0 else (e.args[0] if l == 1 else e.args)
		err_message('Exception raised: "%s"' % msg)



# -*- coding: utf-8 -*-
"""
Serializer module with a slightly improved versions of the Turtle and Pretty XML serializers.

Both are, actually, almost verbatim copy of the RDFLib Turtle and pretty XML serializer module, respectively.
For more detailed on how serializers work and are registered by RDFLib, please refer to the RDFLib 
descriptions and source code.

The differences, v.a.v. the original, for the Pretty XML serializer:
 - there were bugs in the original in case an rdf:List was generated with Literal content
 - added a pretty print branch for Seq/Alt/Bag
 - removal of the CDATA sections when generating an XML Literal (it is a cautionary support in RDFLib but, because XML literals can be generated under very controlled circumstances only in the case of RDFa, it is unnecessary). 
 - use the character " instead of ' for namespace declarations (to make it uniform with the way attributes are handled
 - increased the initial depth for nesting (to 8)
	
The differences, v.a.v. the original, for the Turtle serializer:
 - there was a bug in the syntax of the @prefix setting
 - the original Turtle serializer insisted on creating prefixes for all URI references, ending up with a large number of fairly unnecessary prefixes that made the output a bit unreadable. This has been changed so that only the 'registered' prefixes are used, ie, those that the original RDFa source contains
 - the original Turtle had a bug and did not generate the shorhands for lists
 - changed the indentation rules for anonymous ('squared') blank nodes	
	
"""


# -*- coding: utf-8 -*-
"""
Wrapper around RDFLib's Graph object to provide additional or replacement serializers. The issue is that, in RDFLib 2.X, the turtle and the RDF/XML serialization both have some issues (bugs and ugly output). As a result, the package’s own serializers should be registered and used. On the other hand, in RDFLib 3.X this becomes unnecessary, it is better to keep to the library’s own version. This wrapper provides a subclass of RDFLib’s Graph overriding the serialize method to register, if necessary, a different serializer and use that one.

@summary: Shell around RDLib's Graph
@organization: U{World Wide Web Consortium<http://www.w3.org>}
@author: U{Ivan Herman<a href="http://www.w3.org/People/Ivan/">}
@license: This software is available for use under the U{W3C® SOFTWARE NOTICE AND LICENSE<href="http://www.w3.org/Consortium/Legal/2002/copyright-software-20021231">}
@requires: U{Ordered Dictionary (odict)<http://dev.pocoo.org/hg/sandbox/raw-file/tip/odict.py>}, needed only for the JSON-LD serialization if Python 2.6 or lower is used (Python 2.7 has a built in ordered list module). It is included in the distribution
@requires: U{simplejson package by Bob Ippolito<http://undefined.org/python/#simplejson>}, needed only for the JSON-LD serailization if Python 2.5 or lower is used (Python 2.6 has a json implementation included in the distribution). 
"""

"""
$Id: __init__.py,v 1.4 2013-10-16 11:49:32 ivan Exp $ $Date: 2013-10-16 11:49:32 $

"""

import rdflib
if rdflib.__version__ >= "3.0.0" :
	from rdflib	import Graph
else :
	from rdflib.Graph import Graph
from rdflib	import Namespace

_xml_serializer_name	= "my-rdfxml"
_turtle_serializer_name	= "my-turtle"
_json_serializer_name	= "my-json-ld"

# import rdflib_jsonld
# from rdflib_jsonld.serializer import JsonLDSerializer

try:
	from cStringIO import StringIO
except ImportError:
	if rdflib.__version__ >= "3.0.0" :
		from io import StringIO
	else :
		from StringIO import StringIO

#########################################################################################################
class MyGraph(Graph) :
	"""
	Wrapper around RDFLib's Graph object. The issue is that the serializers in RDFLib are buggy:-(
	
	In RDFLib 2.X both the Turtle and the RDF/XML serializations have issues (bugs and ugly output). In RDFLib 3.X
	the Turtle serialization seems to be fine, but the RDF/XML has problems:-(
	
	This wrapper provides a subclass of RDFLib’s Graph overriding the serialize method to register,
	if necessary, a different serializer and use that one.

	@cvar xml_serializer_registered_2: flag to avoid duplicate registration for RDF/XML for rdflib 2.*
	@type xml_serializer_registered_2: boolean
	@cvar xml_serializer_registered_3: flag to avoid duplicate registration for RDF/XML for rdflib 3.*
	@type xml_serializer_registered_3: boolean
	@cvar json_serializer_registered: flag to avoid duplicate registration for JSON-LD for rdflib 3.*
	@type json_serializer_registered: boolean
	@cvar turtle_serializer_registered_2: flag to avoid duplicate registration for Turtle for rdflib 2.*
	@type turtle_serializer_registered_2: boolean
	"""
	xml_serializer_registered_2		= False
	xml_serializer_registered_3		= False
	turtle_serializer_registered_2	= False
	json_serializer_registered      = False
	
	def __init__(self) :
		Graph.__init__(self)

	def _register_XML_serializer_3(self) :
		"""The default XML Serializer of RDFLib 3.X is buggy, mainly when handling lists. An L{own version<serializers.prettyXMLserializer_3>} is
		registered in RDFlib and used in the rest of the package. 
		"""
		if not MyGraph.xml_serializer_registered_3 :
			from rdflib.plugin import register
			from rdflib.serializer import Serializer
			if rdflib.__version__ > "3.1.0" :
				register(_xml_serializer_name, Serializer,
						 "pyRdfaExtras.serializers.prettyXMLserializer_3_2", "PrettyXMLSerializer")
			else :
				register(_xml_serializer_name, Serializer,
						 "pyRdfaExtras.serializers.prettyXMLserializer_3", "PrettyXMLSerializer")
			MyGraph.xml_serializer_registered_3 = True

	def _register_JSON_serializer_3(self) :
		"""JSON LD serializer 
		"""
		if not MyGraph.json_serializer_registered :
			from rdflib.plugin import register
			from rdflib.serializer import Serializer
			try :
				from rdflib_jsonld.serializer import JsonLDSerializer
				register(_json_serializer_name, Serializer,"rdflib_jsonld.serializer", "JsonLDSerializer")
			except :
				register(_json_serializer_name, Serializer,"pyRdfaExtras.serializers.jsonserializer", "JsonSerializer")
			MyGraph.json_serializer_registered = True

	def _register_XML_serializer_2(self) :
		"""The default XML Serializer of RDFLib 2.X is buggy, mainly when handling lists.
		An L{own version<serializers.prettyXMLserializer>} is
		registered in RDFlib and used in the rest of the package. This is not used for RDFLib 3.X.
		"""
		if not MyGraph.xml_serializer_registered_2 :
			from rdflib.plugin import register
			from rdflib.syntax import serializer, serializers
			register(_xml_serializer_name, serializers.Serializer,
					 "pyRdfaExtras.serializers.prettyXMLserializer", "PrettyXMLSerializer")
			MyGraph.xml_serializer_registered_2 = True

	def _register_Turtle_serializer_2(self) :
		"""The default Turtle Serializers of RDFLib 2.X is buggy and not very nice as far as the output is concerned.
		An L{own version<serializers.TurtleSerializer>} is registered in RDFLib and used in the rest of the package.
		This is not used for RDFLib 3.X.
		"""
		if not MyGraph.turtle_serializer_registered_2 :
			from rdflib.plugin import register
			from rdflib.syntax import serializer, serializers
			register(_turtle_serializer_name, serializers.Serializer,
					 "pyRdfaExtras.serializers.turtleserializer", "TurtleSerializer")
			MyGraph.turtle_serialzier_registered_2 = True
			
	def add(self, t) :
		s,p,o = t
		"""Overriding the Graph's add method to filter out triples with possible None values. It may happen
		in case, for example, a host language is not properly set up for the distiller"""
		if s == None or p == None or o == None :
			return
		else :
			Graph.add(self, (s,p,o))
		
	def serialize(self, format = "xml") :
		"""Overriding the Graph's serialize method to adjust the output format"""
		if rdflib.__version__ >= "3.0.0" :
			# this is the easy case
			if format == "xml" or format == "pretty-xml" :
				self._register_XML_serializer_3()
				return Graph.serialize(self, format=_xml_serializer_name)
			elif format == "json-ld" or format == "json" :
				# The new version of the serialziers in RDFLib 3.2.X require this extra round...
				# I do not have the patience of working out why that is so.
				self._register_JSON_serializer_3()
				stream = StringIO()
				Graph.serialize(self, format=_json_serializer_name, destination = stream, auto_compact = True, indent = 4)
				return stream.getvalue()
			elif format == "nt" :
				return Graph.serialize(self, format="nt")
			elif format == "n3" or format == "turtle" :
				retval =""
				return Graph.serialize(self, format="turtle")
		else :
			if format == "xml" or format == "pretty-xml" :
				self._register_XML_serializer_2()
				return Graph.serialize(self, format=_xml_serializer_name)
			elif format == "nt" :
				return Graph.serialize(self, format="nt")
			elif format == "n3" or format == "turtle" :
				self._register_Turtle_serializer_2()
				return Graph.serialize(self, format=_turtle_serializer_name)





"""

Serializer for RDF/XML. Note that this is a slightly modified version of RDFLib's PrettyXMLSerializer module; the
original version had some bugs (mainly in handling collections) hence this local version.
"""


from rdflib import RDF, RDFS

from rdflib	   import URIRef, Literal, BNode
from rdflib.util  import first, uniq, more_than
from rdflib.Graph import Seq

from rdflib.syntax.serializers	import Serializer
from pyRdfa.serializers.XMLWriter import XMLWriter

XMLLANG = "http://www.w3.org/XML/1998/namespacelang"


# TODO:
def fix(val):
	"strip off _: from nodeIDs... as they are not valid NCNames"
	if val.startswith("_:"):
		return val[2:]
	else:
		return val

class PrettyXMLSerializer(Serializer):

	def __init__(self, store, max_depth=8):
		super(PrettyXMLSerializer, self).__init__(store)
		
		# collecting information on lists, for a proper and pretty printing of those...
		self.list_heads        = []
		self.list_really_heads = []
		for l in self.store.subjects(RDF.first,None) :
			# yep, l is part of a list. But we want to retain exclusively those that are heads and not
			# internal list elements
			# furthermore: we want list heads which have only non-Literal elements. This is the unfortunate
			# consequence of the RDF/XML syntax...:-(
			items = [item for item in store.items(l)]
			if True not in [ isinstance(item,Literal) for item in items ] :
				references = [ r for r in self.store.subjects(RDF.rest,l) ]
				self.list_heads.append(l)
				if len(references) == 0 :
					self.list_really_heads.append(l)

	def serialize(self, stream, base=None, encoding=None, **args):
		self.__serialized = {}

		# "Internal" list elements should be forgotten...
		for l in self.list_heads :
			if l not in self.list_really_heads :
				self.__serialized[l] = 1

		store = self.store
		self.base = base
		#self.max_depth = args.get("max_depth", 3)
		self.max_depth = 8

		self.nm = nm = store.namespace_manager
		self.writer = writer = XMLWriter(stream, nm, encoding)

		writer.push(RDF.RDF)

		# This checks whether a predicate appears with an unallowed URI and creates
		# an artificial namespace if the answer is yes...		
		for predicate in uniq(store.predicates()):
			prefix, namespace, name = nm.compute_qname(predicate)
		# The same for types, that can appear as XML element names
		for subj,types in uniq(store.subject_objects(RDF.type)) :
			prefix, namespace, name = nm.compute_qname(types)
	
		ns_list= list(self.store.namespaces())
		ns_list.sort()
		writer.namespaces(ns_list)

		# Write out subjects that can not be inline
		for subject in store.subjects():
			if (None, None, subject) in store:
				if (subject, None, subject) in store:
					self.subject(subject, 1)
			else:
				self.subject(subject, 1)

		# write out anything that has not yet been reached
		for subject in store.subjects():
			self.subject(subject, 1)

		writer.pop(RDF.RDF)

		# Set to None so that the memory can get garbage collected.
		self.__serialized = None

	def subject(self, subject, depth=1):
		store  = self.store
		writer = self.writer
		if not subject in self.__serialized:
			self.__serialized[subject] = 1
			type = first(store.objects(subject, RDF.type))
			try:
				self.nm.qname(type)
			except:
				type = None
			element = type or RDF.Description
			writer.push(element)
			if isinstance(subject, BNode):
				def subj_as_obj_more_than(ceil):
					return more_than(store.triples((None, None, subject)), ceil)
				if (depth == 1 and subj_as_obj_more_than(0)) or subj_as_obj_more_than(1):
					writer.attribute(RDF.nodeID, fix(subject))
			else:
				writer.attribute(RDF.about, self.relativize(subject))
			if (subject, None, None) in store:
				for predicate, object in store.predicate_objects(subject):
					if not (predicate==RDF.type and object==type):
						self.predicate(predicate, object, depth+1)
			writer.pop(element)

	def predicate(self, predicate, object, depth=1):
		def defaultCase() :
			if depth <= self.max_depth:
				self.subject(object, depth+1)
			elif isinstance(object, BNode):
				writer.attribute(RDF.nodeID, fix(object))
			else:
				writer.attribute(RDF.resource, self.relativize(object))
		#####
		
		writer = self.writer
		store = self.store
		writer.push(predicate)
		
		if isinstance(object, Literal):
			attributes = ""
			to_escape = True
			if object.language:
				writer.attribute(XMLLANG, object.language)
			if object.datatype:
				if ("%s" % object.datatype) == "http://www.w3.org/1999/02/22-rdf-syntax-ns#XMLLiteral" :
					writer.attribute(RDF.parseType, "Literal")
					to_escape = False
				else :
					writer.attribute(RDF.datatype, object.datatype)
			writer.text(object,to_escape)
		elif object in self.list_heads :
			items = [item for item in store.items(object)]
			if True not in [ isinstance(item,Literal) for item in items ] :
				# This is a kosher list that could be handled with the Collection parse type trick
				self.__serialized[object] = 1
				if object in self.list_really_heads :
					collection = object
					self.__serialized[object] = 1
					writer.attribute(RDF.parseType, "Collection")
					for item in items :
						if item in self.__serialized :
							# bugger; already done somewhere else... :-)
							writer.push(RDF.Description)
							if isinstance(item, BNode):
								if more_than(store.triples((None, None, item)), 0):
									writer.attribute(RDF.nodeID, fix(item))
							else:
								writer.attribute(RDF.about, self.relativize(item))
							writer.pop(RDF.Description)
						else :
							self.subject(item)
							self.__serialized[item] = 1
			else :
				defaultCase()
		elif object in self.__serialized or not (object, None, None) in store:
			if isinstance(object, BNode):
				if more_than(store.triples((None, None, object)), 0):
					writer.attribute(RDF.nodeID, fix(object))
			else:
				writer.attribute(RDF.resource, self.relativize(object))
		elif (object,RDF.type,RDF.Seq) in store or (object,RDF.type,RDF.Bag) in store or (object,RDF.type,RDF.Alt) in store :
			seq = Seq(store,object)
			self.__serialized[object] = 1
			if (object,RDF.type,RDF.Seq) in store :
				typ = RDF.Seq
			elif (object,RDF.type,RDF.Alt) in store :
				typ = RDF.Alt
			else :
				typ = RDF.Bag
			writer.push(typ)
			for item in seq :
				self.predicate(RDF.li,item,depth+1)
			writer.pop(typ)
		else :
			defaultCase()
		writer.pop(predicate)




from setuptools import setup

install_requires=[
    "rdflib",
    "html5lib",
]

setup(name="pyRdfa3",
      description="pyRdfa Libray",
      version="3.5.3",
      author="Ivan Herman",
      author_email="ivan@w3.org",
      maintainer="Ivan Herman",
      maintainer_email="ivan@w3.org",
      install_requires=install_requires,
      packages=['pyRdfa',
				'pyRdfa.transform',
				'pyRdfa.extras',
				'pyRdfa.rdfs',
				'pyRdfa.host',
				'pyRdfaExtras',
				'pyRdfaExtras.extras',
				'pyRdfaExtras.serializers'
				],
    entry_points={
        'rdf.plugins.parser': [
            'rdfa = pyRdfa.rdflibparsers:RDFaParser',
            'rdfa1.0 = pyRdfa.rdflibparsers:RDFa10Parser',
            'rdfa1.1 = pyRdfa.rdflibparsers:RDFaParser',
            'application/svg+xml = pyRdfa.rdflibparsers:RDFaParser',
            'application/xhtml+xml = pyRdfa.rdflibparsers:RDFaParser',
            'hturtle = pyRdfa.rdflibparsers:HTurtleParser'
            'html = pyRdfa.rdflibparsers:StructuredDataParser'
            'html/text = pyRdfa.rdflibparsers:StructuredDataParser'
        ],
    }

)



"""

Serializer for Turtle. This is a slightly modified version of RDFLib's TurtleSerializer module; the
original version had some bugs (in defining prefixes), and the overall output look has also been slightly improved.

"""

import urlparse
from xml.sax.saxutils 			import escape, quoteattr

from rdflib.BNode 				import BNode
from rdflib.Literal 			import Literal
from rdflib.URIRef 				import URIRef
from rdflib.syntax.xml_names 	import split_uri 

from rdflib.syntax.serializers.RecursiveSerializer import RecursiveSerializer
from rdflib.exceptions import Error

from rdflib import RDF, RDFS

XSD = "http://www.w3.org/2001/XMLSchema#"

SUBJECT = 0
VERB = 1
OBJECT = 2

def _Literal_n3(lit,store):
	language = lit.language
	datatype = lit.datatype
	# Modifying the datatype to use a prefixed value rather than the whole URI
	# One has to be a bit careful, because user defined datatypes are also possible
	# The rest is a copy of the original method took over from Literal
	if datatype :
		datatype = store.namespace_manager.normalizeUri(datatype)

	# TODO: We could also chose quotes based on the quotes appearing in the string, i.e. '"' and "'" ...

	# which is nicer?
	# if lit.find("\"")!=-1 or lit.find("'")!=-1 or lit.find("\n")!=-1:
	if lit.find("\n") != -1:
		# Triple quote this string.
		encoded = lit.replace('\\', '\\\\')
		if lit.find('"""')!=-1: 
			# is this ok?
			encoded=encoded.replace('"""','\\"""')
		if encoded.endswith('"'): encoded=encoded[:-1]+"\\\""
		encoded = '"""%s"""' % encoded
	else: 
		encoded = '"%s"' % lit.replace('\n','\\n').replace('\\', '\\\\').replace('"','\\"')
	if language:
		if datatype:    
			return '%s@%s^^%s' % (encoded, language, datatype)
		else:
			return '%s@%s' % (encoded, language)
	else:
		if datatype:
			return '%s^^%s' % (encoded, datatype)
		else:
			return '%s' % encoded
			

class TurtleSerializer(RecursiveSerializer):
	short_name="turtle"
	indentString = "    "
	def __init__(self, store):
		super(TurtleSerializer, self).__init__(store)
		self.reset()
		self.stream = None
		self.store.namespace_manager.bind("xsd","http://www.w3.org/2001/XMLSchema#")

	def reset(self):
		super(TurtleSerializer, self).reset()
		self._shortNames = {}
		self._started = False
	
	def getQName(self, uri):
		if isinstance(uri, URIRef):
			return self.store.namespace_manager.normalizeUri(uri)
		else :
			return None

	def preprocessTriple(self, triple):
		super(TurtleSerializer, self).preprocessTriple(triple)
		p = triple[1]
		if isinstance(p, BNode):
			self._references[p] = self.refCount(p) +1
			
	def label(self, node):
		qname = self.getQName(node)
		if qname is None:
			if isinstance(node,Literal) :
				return _Literal_n3(node,self.store)
			else :
				return node.n3()
		return qname

	def startDocument(self):
		self._started = True
		ns_list= list(self.store.namespaces())
		ns_list.sort()
		if len(ns_list) == 0:
			return
		
		for prefix, uri in ns_list:
			self.write(self.indent() + '@prefix %s: <%s> .\n' % (prefix, uri))
			
		self.write('\n')

	def endDocument(self):
		pass

	def isValidList(self,l): 
		"""Checks if l is a valid RDF list."""			
		return len([ i for i in self.store.items(l) ]) > 0
		
	def doList(self,l):
		while l:
			item = self.store.value(l, RDF.first)
			if item:
				self.path(item, SUBJECT)
				self.write(' ')
				self.subjectDone(l)
			l = self.store.value(l, RDF.rest)
			
	def p_squared(self, node, position):
		if not isinstance(node, BNode) or node in self._serialized or self.refCount(node) > 1 or position == SUBJECT:
			return False

		if self.isValidList(node): 
			# this is a list
			self.write(' ( ')
			self.depth += 2
			self.doList(node)
			self.write(')')
			self.depth -= 2			
		else :
			self.subjectDone(node)
			self.depth += 2
			self.write('\n' + self.indent() + ' [')
			self.predicateList(node)
			self.write('\n' + self.indent() + ' ]')
			self.depth -= 2
		return True

	def p_default(self, node, ignore):
		if ignore == SUBJECT :
			self.write(self.label(node))
		else :
			self.write(" " + self.label(node))
		return True
	
	def path(self, node, position):
		if not (self.p_squared(node, position) or self.p_default(node, position)):
			raise Error("Cannot serialize node '%s'" % (node, ))

	def verb(self, node):
		if node == RDF.type:
			self.write(' a')
		else:
			self.path(node, VERB)
	
	def objectList(self, predicate, objects):
		num = len(objects)
		if num == 0:
			return
		elif num > 3 :
			self.write('\n' + self.indent(2))

		self.path(objects[0], OBJECT)
		for obj in objects[1:]:
			if num > 3 :
				self.write(',\n' + self.indent(2))
			else :
				self.write(",")
			self.path(obj, OBJECT)

	def predicateList(self, subject):
		properties = self.buildPredicateHash(subject)
		propList = self.sortProperties(properties)
		if len(propList) == 0:
			return

		self.verb(propList[0])
		self.objectList(propList[0],properties[propList[0]])
		for predicate in propList[1:]:
			self.write(' ;\n'+self.indent(1))
			self.verb(predicate)
			self.objectList(predicate,properties[predicate])

	def s_squared(self, subject):
		if (self.refCount(subject) > 0) or not isinstance(subject, BNode) :
			return False
		self.write('\n' + self.indent() + " []")
		self.depth += 1
		self.predicateList(subject)
		self.write(' .')
		self.depth -= 1
		#self.write('].')
		return True

	def s_default(self, subject):
		self.write('\n' + self.indent())
		self.path(subject, SUBJECT)
		self.predicateList(subject)
		self.write(' . ')
		return True
	
	def statement(self, subject):
		self.subjectDone(subject)
		if not self.s_squared(subject):
			self.s_default(subject)
			
	def serialize(self, stream, base=None, encoding=None, **args):
		self.reset()
		self.stream = stream
		self.base=base
		
		self.preprocess()
		subjects_list = self.orderSubjects()

		self.startDocument()

		firstTime = True
		for subject in subjects_list:
			if not self.isDone(subject):
				if firstTime:
					firstTime = False
				else:
					self.write('\n')
				self.statement(subject)
		
		self.endDocument()
